/* commandird  - Initial version in CommandIR Utils 0.5
   0.5 - Support multiple i/o pipes, queue signals, send & receive without restarting program
 */

#include "cmdirutils_version.h"
#include <signal.h>
#include "libcmdir.h"
#include "libtcp.h"

int commandird_nodaemon;  // move to libcmdir?  More than us use it?

extern struct commandir_tx_order * ordered_commandir_devices;
extern unsigned char pipeOnly;
extern struct commandir_device * first_commandir_device;
extern int tcpTimingListenersCount;

char unsigned commandir_data_buffer[512];
unsigned char standalone = 0; // NETWORK server, not standalone program

/* Prototypes - remote */
void init_tcp() ;               // libtcp
void check_tcp_connections();   // libtcp

void closePipes();              // libpipes

/* Prototypes - local */
void check_commandir_queue(void) ;
int check_commandir_rec(void) ;
void commandir_disconnect(struct commandir_device * pcd);
void hardware_disconnect(struct commandir_device *a);
void software_disconnects(void);
void check_commandir_add_remove(void);
void display_help(void);

int check_term_signal()
{
  // See if we've received a SIG_TERM
  return 0;
}

char * progress = "-/|\\";
int pnum = 0;

void display_help() 
{
  printf("commandird %s\n\n", utils_version);
  printf(" Daemon for programs to send and receive from CommandIR III's via Pipes and TCP/IP\n\n");

  printf("Usage: commandird [-d] [-v] [-sCOMMANDIR_NAME] [-sCOMMANDIR_NAME2] ...\n\n");

  printf(" Run with -d or -v to setup network and pipe servers.\n");
  printf(" -d - run as a daemon\n");
  printf(" -v - Verbose - do not daemonize; display all activity\n");
  printf(" -sCOMMANDIR_NAME - Use only specific CommandIRs (Firmware >=3.42 required)\n\n");
  printf("TCP/IP Networking: commandird accepts TCP connections on port 65123\n");
  printf("  Simply use commandir_record and commandir_send in network mode (with the -n switch)\n");
  printf("  to use commandird remotely\n\n");
  printf("Pipes: commandird uses two pipes by default:\n");
  printf("  /tmp/commandir_cmd: Create with 'mkfifo /tmp/commandir_cmd'\n");
  printf("    To use this command pipe, send a command string to the pipe. \n");
  printf("    Examples:  \n");
  printf("    echo \"commandir_send -b5 -r 0000 0074 0017 0000 0020 0020 0040 0040 0020 0021 0020 0020 0020 0021 0020 0020 0020 0021 003F 0021 0020 001F 0020 0021 0020 0020 0020\" > /tmp/commandir_rec\n");
  printf("    echo \"commandir_clearmem\" > /tmp/commandir_cmd\n");
  printf("    echo \"commandir_record -g100000\" > /tmp/commandir_cmd\n");
  printf("  commandir_rx: Create with 'mkfifo /tmp/commandir_rx'\n");
  printf("    To use this rx pipe, simply read the pipe to see the hex code data. \n");
  printf("    Example:  \n");
  printf("    cat /tmp/commandir_rx\n");
  return;
}




int main(int argc, char **argv)
{
  int x, myid, noHelp = 0;
  struct commandir_device * pcd;
  int lc;
  int rxed = 0;
  
  for(x=1; x<argc; x++)
  {
    if(argv[x][0] == '-')
    {
      switch(argv[x][1])
      {
        case 'd':
          // Trigger to go ahead and do something other than display help
          noHelp = 1;
          break;
        case 's':
          // (another) Named CommandIR - search for this one - don't we want to do this after the fork()?
          validate_param_s(argv[x]);
          reserveCommandIR_By_Name(&argv[x][2]);
          break;
        case 'v':
          // Verbose - do not daemonize 
          commandird_nodaemon = 1;
          noHelp = 1;
          break;
        default:
          printf("Unknown command line argument %c\n", argv[x][1]);
          display_help();
          return -1;
      }
    } 
  }

  if(noHelp == 0) {
    display_help();
    exit(0);
  }

#ifndef WIN
  if(commandird_nodaemon == 0) 
  {
    myid = fork();
    if(myid) {
      return 0; // Parent process exits
    }
  }
  
  signal(SIGPIPE, SIG_IGN);
#endif

  tcpRxListenersCount = 0;
  
  hardware_scan();  
  detect_tx_per_commandir();
  
  if(ordered_commandir_devices) {
    hardware_setorder_by_pdata_order();
    connectReservedCommandIRs();  // Only NAMED CommandIRs
    disconnectUnneededCommandIRs();
  } else {
//    hardware_setorder();  // ALL CommandIRs
//		printf("commandird - setting up order\n");
		clear_order();
		hardware_setorder_by_pdata_order();
  }
  
  // Init TCP
  init_tcp();

  pipeOnly = commandird_nodaemon; 

  // Main loop 
  for(;;)
  {
    // Check for hardware changed - CommandIRs added/removed?
    check_commandir_add_remove();
    if(tcpRxListenersCount + tcpTimingListenersCount > 0) 
    {
      // Check commandirs for incoming RX data
      rxed = check_commandir_rec();  // Any new rec data, if we have active_rec clients?
    }

    check_tcp_connections();  // If connected, returns; if not, waits for connection

    check_tcp_incoming(); // recv to check for data - HANG?

#ifndef WIN
    check_pipe_statuses();
    if(check_term_signal())      // Any signals for us to terminate?  On Windows?
      break;  // We're done
#endif

    check_commandir_queue();  // Anything we have queued to send to CommandIRs?
    
//    printf("%c\n", progress[pnum++]);
    if(pnum>=4) pnum = 0;
    if(rxed == 0)
#ifdef WIN
      Sleep(100);
#else
      usleep(100000); // otherwise, get it to clients QUICK
#endif
  }
  
#ifndef WIN
  closePipes();
#endif
  return 0;
}



void commandir_iii_update_status(struct commandir_device *cd)
{
	int receive_status;
	struct commandirIII_status *sptr;

	// Ready the first status that will tell us all the above info
	receive_status = usb_bulk_read(cd->cmdir_udev, 1,	// endpoint 1
				       (char *)commandir_data_buffer, cd->endpoint_max[1], 1500);

	if (receive_status == 8) {
		// Update the commandir_device with what we just received.
		sptr = (struct commandirIII_status *)commandir_data_buffer;
		cd->num_transmitters = (sptr->tx_status & 0x1F) + 1;
		cd->num_receivers = (sptr->rx_status & 0x60) >> 5;
		cd->tx_jack_sense =
		    sptr->jack_status[3] * 0x01000000 + sptr->jack_status[2] * 0x010000 +
		    sptr->jack_status[1] * 0x0100 + sptr->jack_status[0];
		cd->rx_jack_sense = sptr->rx_status & 0x03;
		cd->commandir_tx_available = (sptr->tx_status & 0x80 ? 1024 : 0);
		cd->rx_data_available = sptr->rx_status & 0x80;
		cd->hw_revision = sptr->versionByte >> 5;	// 3 high bits
		cd->hw_subversion = sptr->versionByte & 0x1F;	// 5 low bits
	}
}

/* Either replace tx_signal with sendir, or convert */
int sendToCommandIRifSpace(struct commandir_device * pcd, struct tx_signal * s)
{
  struct sendir sir;
  // Refresh it's status for latest availability
  commandir_iii_update_status(pcd);
  if(pcd->commandir_tx_available > s->raw_signal_len)
  {
    // convert tx_signal to sendir
    // commandir_send_specific wants a struct sendir
    sir.buffer = s->raw_signal;
    sir.byteCount = s->raw_signal_len;
    sir.emitterMask = s->raw_signal_tx_bitmask;
    commandir_send_specific(&sir, pcd);  // doesn't clear malloc's
    return 1;
  }
  return 0;
}


void check_commandir_queue() 
{
  struct tx_signal * s, * olds;
  struct commandir_device * t;
  
  if(first_commandir_device == NULL) return;
  
  // Any queue data?
  for(t = first_commandir_device; 
      t != NULL; 
      t=t->next_commandir_device)
  {
    if(t->next_tx_signal == NULL) continue;
    s = t->next_tx_signal;

	olds = s;
    while(s != NULL) 
    {
      // Is there space on this CommandIR to move this signal to 
      if(sendToCommandIRifSpace(t, s)==0)
      {
        break;  // break out of the while; if there is no space for this signal, we cannot send the next yet.
      }
      // Otherwise, we clean up this tx_signal:
      free(s->raw_signal);
      
      if(t->next_tx_signal == s)
        t->next_tx_signal = s->next;

      if(olds != NULL) 
      {
        olds->next = s->next;
      }
      free(s);
    }    
  }
}


int check_commandir_rec()
{
  struct commandir_device * pcd;
  int read_retval;
  char incoming_packet[512];  // 64 should be enough
  
  int totalRec = 0;
  int thisDevNum, flagged_disconnect = 0;
  
  if(first_commandir_device == NULL)  return 0;
  
  for(pcd = first_commandir_device; pcd != NULL; pcd = pcd->next_commandir_device) 
  {
    thisDevNum = 1;
    while(thisDevNum > 0) {
      thisDevNum = commandir_rec(pcd->cmdir_udev);
      if(thisDevNum < 0) {
        pcd->flag_disconnect = 1;
        flagged_disconnect++;
      } else {
        totalRec += thisDevNum;
      }
    }
  }
  if(flagged_disconnect > 0)
    disconnectUnneededCommandIRs();
  return totalRec;
}



